#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <stdarg.h>

#if defined( MPATROL )
#include <mpatrol.h>
#endif /* MPATROL */

/*
 * TODO:
 * this program should be modified to handle wide characters.
 */

/*
 * these three DEFAULT_ defines do not need to be changed ever, as
 * they are only initial values.
 * LINES and BUFSIZE will grow dynamically as needed, during program
 * execution; if you want to change the defaults however, define them
 * in the Makefile.
 * TABWIDTH can be changed through runtime option -t.
 */

#if !defined( DEFAULT_LINES )
#define DEFAULT_LINES		128		/* no need to change, will grow dynamically */
#endif /* DEFAULT_LINES */

#if !defined( DEFAULT_BUFSIZE )
#define DEFAULT_BUFSIZE		(size_t)4096	/* no need to change, will grow as needed */
#endif /* DEFAULT_BUFSIZE */

#if !defined( DEFAULT_TABWIDTH )
#define DEFAULT_TABWIDTH	8		/* can be set with option -t */
#endif /* DEFAULT_TABWIDTH */

#if !defined( BUILD_DATE )
#define BUILD_DATE	"unknown"
#endif /* BUILD_DATE */

const char author[]="bjdouma@xs4all.nl";

typedef struct
{
	int   len;
	char *str;	/* to be malloc'ed */
	char *curp;	/* current pointer into str */
	int   vindex;	/* current virtual index including total space added from expanded tabs */
}
 LINE;

#define EXIT			1
#define CONT			0

#define TAB			(char)'\t'
#define SPC			(char)' '

#define info( fmt... )		fprintf( stderr, "> "fmt )
#define ZAP( p )		{ if (p!=NULL){ free( p ); p=NULL; } }		/* free a pointer and set it to NULL */

static void  print_usage( void );
static void  get_options( int argc, char **argv );
static int   check_numeric_optarg( void );
static void  process_files( int argc, char *argv[] );
static void  read_file( FILE *fi );
static void  tilt_file( void );
static void  clear_lines( void );
static char *string_alloc( const char *bufname, char *old_buf, size_t len );
static LINE *lines_alloc( LINE *old_lines, size_t nlines_old, size_t nlines_new );

char **args;
LINE *lines;
char *linebuf=NULL;	/* buffer for one input line */
int nlines;
int longest_len=0;
size_t lines_allocated=0;
size_t linebuf_allocated=0;

/* command-line options */
int tabwidth=DEFAULT_TABWIDTH;
int usage=0;
int verbose=0;
int strip_whitespace=0;

/*---------------------------------------------------------------------------*/
static void error( int act, int err, const char *fmt, ... )
{
	static char msgbuf[512];
	va_list argptr;

	va_start( argptr, fmt );

	vsprintf( msgbuf, fmt, argptr );

	fprintf( stderr, "%s: %s", args[0], msgbuf );
	if (err)
		fprintf( stderr, ": %s (errno=%d)\n", strerror( err ), err );
	else
		fprintf( stderr, "\n" );

	fflush( stdout );
	fflush( stderr );

	if (act==EXIT)
		exit( 1 );

	va_end( argptr );
}

/*---------------------------------------------------------------------------*/
static char *string_alloc( const char *bufname, char *old_buf, size_t len )
{
	if ((old_buf=(char *)realloc( old_buf, len*sizeof( char ) ))==NULL)
		error( EXIT, errno, bufname );

	if (verbose)
		info( "allocated %d bytes for '%s'\n", len*sizeof( char ), bufname );

	return( old_buf );
}

/*---------------------------------------------------------------------------*/
static LINE *lines_alloc( LINE *old_lines, size_t nlines_old, size_t nlines_new )
{
	register unsigned int i;

	if ((old_lines=(LINE *)realloc( old_lines, nlines_new*sizeof( LINE ) ))==NULL)
		error( EXIT, errno, "lines_alloc" );

	/*
	 * malloc returns uninitialized memory... we initialize here, in order
	 * to be able to free any lines[].str pointers that are not NULL
	 * after a file has been tilted (see clear_lines()).
	 */
	for ( i=nlines_old; i<nlines_new; i++ )
		old_lines[i].str = NULL;

	if (verbose)
		info( "allocated %d line structures (totalling %d bytes)\n", nlines_new, nlines_new*sizeof( LINE ) );

	return( old_lines );
}

/*---------------------------------------------------------------------------*/
/*
 * free all string pointers allocated _within_ the line structures,
 * since we can not reuse these for a possible next file.
 *
 */

static void clear_lines( void )
{
	register int i;

	for ( i=0; i<(int)lines_allocated; i++ )
		ZAP( lines[i].str );
}

/*---------------------------------------------------------------------------*/
/*
 * read in all lines from the file and determine
 * the longest line length, and the number of lines.
 *
 */

static void read_file( FILE *fi )
{
	register int len, done;
	int missing_newline=0;

	if (verbose)
		info( "reading file ...\n" );

	nlines = 0;
	longest_len = 0;
	done = 0;

	/*
	 * allocate a buffer for fgets to read input lines; in case it's too small,
	 * it will grow dynamically.
	 */
	linebuf = string_alloc( "linebuf", linebuf, linebuf_allocated=DEFAULT_BUFSIZE );

	while (fgets( (char *)linebuf+done, (int)linebuf_allocated-done, fi ))
	{
		done = 0;
		missing_newline = 0;
		len = strlen( linebuf );

		/* should never happen, since linebuf grows dynamically */
		if ((size_t)len>linebuf_allocated-1)
			error( EXIT, errno, "bufsize too small (line=%d,len=%d,bufsize=%d)", nlines+1, len, linebuf_allocated-1 );

		if (len>0)
		{
			if (linebuf[len-1]=='\n')
			{
				/* strip off NL and/or msdos CR */
				while (len>0 && (linebuf[len-1]=='\n' || linebuf[len-1]=='\r'))
					linebuf[--len] = 0;
			}
			else
			{
				/*
				 * fgets() stores a NUL at the end of its buffer; if fgets() got the line
				 * in its entirety, then there should be a '\n' before that NUL, except
				 * in the case when fgets() reached EOF /and/ there was no NL at EOF.
				 */

				int next_char;

				while (len>0 && (linebuf[len-1]=='\n' || linebuf[len-1]=='\r'))
					linebuf[--len] = 0;

				if ((next_char=fgetc( fi ))==EOF)
					missing_newline = 1;
				else
				if (next_char!='\n')
				{
					if (verbose)
						info( "oops, line %d too long for linebuf, could only read %d characters, expanding linebuf\n", nlines+1, len+1 );

					linebuf = string_alloc( "linebuf", linebuf, linebuf_allocated*=2 );
					linebuf[len] = next_char;
					done = len+1;			/* +1 because it includes next_char */

					continue;
				}
			}
		}

		if (len>longest_len)
			longest_len = len;

		if ((size_t)nlines>=lines_allocated)
		{
			lines = lines_alloc( lines, lines_allocated, lines_allocated*2 );
			lines_allocated *= 2;
		}

		lines[nlines].len = len;
		lines[nlines].str = NULL;
		lines[nlines].str = string_alloc( "lines[].str", lines[nlines].str, (size_t)len+1 );
		strncpy( lines[nlines].str, linebuf, (size_t)len+1 );
		lines[nlines].curp = lines[nlines].str;
		lines[nlines].vindex = 0;

		nlines++;
		done = 0;
	}

	if (verbose)
	{
		register int i;

		info( "number of lines = %d\n", nlines );
		info( "longest length = %d\n", longest_len );
		info( "dumping all lines:\n" );
		for ( i=0; i<nlines; i++ )
			fprintf( stderr, "%s\n", lines[i].str );
		if (missing_newline)
			info( "missing newline at EOF -- no problem.\n" );
	}

	ZAP( linebuf );
}

/*---------------------------------------------------------------------------*/
/*
 * tilt file (expand tabs along the way).
 */

static void tilt_file( void )
{
	register int i;		/* i is the real index, vindex is the virtual index */
	register LINE *L;
	register char *o;	/* pointer into outbuf */
	int nspaces;		/* spaces needed to expand a tab */
	char *eol;		/* position of EOL in outbuf */
	char *outbuf=NULL;	/* buffer for printing output lines */

	outbuf = string_alloc( "outbuf", outbuf, (size_t)nlines+1 );

	if (verbose)
		info( "dumping lines tilted 90 degrees clockwise:\n" );

	/*
	 * note: longest_len can and may be adjusted from inside the loop
	 */
	for ( i=0; i<longest_len; i++ )
	{
		for ( o=eol=outbuf, L=&lines[nlines-1]; o<outbuf+nlines; o++, L-- )
		{
			if (i>=L->len)
				*o = SPC;
			else
				*o = *(L->curp);

			if (*o==TAB && tabwidth>0)
			{
				*o = SPC;

				if (i==L->vindex)		/* first time we encounter this tab */
				{
					nspaces = tabwidth - 1 -  ((L->vindex) & (tabwidth - 1));

					/* increase virtual index for this line */
					L->vindex += nspaces + 1;

					/* adjust the line's length */
					L->len += nspaces;

					/* also maybe adjust longest_len */
					if (L->len>longest_len)
						longest_len = L->len;
				}

				if (i==L->vindex-1)		/* expanding tab about to finish */
					L->curp++;
			}
			else
			{
				L->curp++;
				L->vindex++;
			}

			/* if alphanumeric character, push ahead eol */
			if (*o!=SPC && *o!=TAB)
				eol = o+1;
		}

		if (strip_whitespace)
			*eol = '\0';
		else
			*o = '\0';

		fprintf( stdout, "%s\n", outbuf );
	}

	fflush( stdout );

	ZAP( outbuf );
}

/*---------------------------------------------------------------------------*/
#define _GNU_SOURCE
#include <getopt.h>

#define OPT_HELP		"help"
#define OPT_STRIP_WHITESPACE	"strip-whitespace"
#define OPT_TABWIDTH		"tabwidth"
#define OPT_VERBOSE		"verbose"

static struct option long_options[]=
{
	/* name			has_arg			flag  val */
	/* ----------------	-----------------	----  --- */
	{ OPT_HELP,		no_argument,		NULL, 'h' },
	{ OPT_STRIP_WHITESPACE,	no_argument,		NULL, 's' },
	{ OPT_TABWIDTH,		required_argument,	NULL, 't' },
	{ OPT_VERBOSE,		no_argument,		NULL, 'v' },
	{ NULL,			0,			NULL,  0  }
};

#define INVALID_OPTARG_MSG	"invalid argument '%s' to option -%c"

static void get_options( int argc, char **argv )
{
	int c, option_index=0;

	while (1)
	{
		if ((c=getopt_long( argc, argv, "hst:v", long_options, &option_index ))==-1)
			break;

		switch (c)
		{
			case 'h':	usage = 1;
					break;

			case 's':	strip_whitespace = 1;
					break;

			case 't':	if (!check_numeric_optarg() || ((tabwidth=(size_t)atoi( optarg ))<=0))
						error( EXIT, errno, INVALID_OPTARG_MSG, optarg, c );
					break;

			case 'v':	verbose = 1;
					break;

			case ':':	error( EXIT, errno, "missing option parameter" );
					break;

			case '?':	error( EXIT, errno, "ambiguous option" );
					break;

			default:	error( EXIT, errno, "unrecognized option" );
					break;
		}
	}

	if (verbose)
	{
		info( "%s %s\n\n", args[0], VERSION );
		info( "current options:\n" );
		info( OPT_VERBOSE"          = true\n" );
		info( OPT_STRIP_WHITESPACE" = %s\n", strip_whitespace?"true":"false" );
		info( OPT_TABWIDTH"         = %d\n", tabwidth );
		info( "\n" );
	}
}

/*---------------------------------------------------------------------------*/
static void print_usage( void )
{
	fprintf( stderr, "%s %s (%s)\n", args[0], VERSION, author );
	fprintf( stderr, "(built: %s)\n", BUILD_DATE );
	fprintf( stderr, "program/filter to tilt file contents at character level 90 degrees clockwise\n\n" );
	fprintf( stderr, "usage:\n" );
	fprintf( stderr, "tilt [OPTION]... [FILE]...\n\n" );
	fprintf( stderr, "-h, --"OPT_HELP"	       show this help screen and exit.\n" );
	fprintf( stderr, "-s, --"OPT_STRIP_WHITESPACE" strip off trailing whitespace on output lines (default: keep as is).\n" );
	fprintf( stderr, "-t, --"OPT_TABWIDTH"=NUM     expand tabs to a width of NUM spaces (default: %d).\n", DEFAULT_TABWIDTH );
	fprintf( stderr, "-v, --"OPT_VERBOSE"	       report some data to stderr while processing files.\n" );
}

/*---------------------------------------------------------------------------*/
static int check_numeric_optarg( void )
{
	register char *p;

	for ( p=optarg; *p; p++ )
		if (*p<'0' || *p>'9')
			return( 0 );
	return( 1 );
}

/*---------------------------------------------------------------------------*/
int main( int argc, char **argv )
{
	args = argv;

	get_options( argc, argv );

	if (usage)
	{
		print_usage();
		exit( 1 );
	}
	else
		process_files( argc, argv );

	if (verbose)
		info( "done.\n" );

	/* all done */
	return( 0 );
}

/*---------------------------------------------------------------------------*/
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifndef _POSIX_SOURCE
#include <sys/ioctl.h>
#endif

static char *infile;		/* name of input file; may be "-" */
static int input_desc;		/* descriptor on which input file is open */
static int output_desc;		/* descriptor on which output file is open; always is 1 */
static FILE *input_stream;

#define is_stdin(filename)	((filename)[0]=='-' && (filename)[1]=='\0')
#define STDIN			0
#define STDOUT			1

/*
 * most of this snatched from old `cat' sources (gnu textutils)
 * courtesy of Torbjorn Granlund and Richard M. Stallman;
 * possible bugs courtesy me -- bjd.
 */

static void process_files( int argc, char *argv[] )
{
	int argind;		/* index in argv to processed argument */
	dev_t out_dev=0;	/* device number of the output (file or whatever) */
	dev_t out_ino=0;	/* i-node number of the output */
	int check_redirection;	/* nonzero if the output file should not be the same as any input file */
	int have_read_stdin=0;	/* nonzero if we have ever read standard input */
	struct stat stat_buf, *input_stat=&stat_buf, *output_stat=&stat_buf;

	if (verbose)
		info( "processing files...\n" );

	output_desc = STDOUT;

	/* get device and i-node */
	if (fstat( output_desc, output_stat )<0)
		error( EXIT, errno, "stdout" );
	if (S_ISREG( output_stat->st_mode ))
	{
		out_dev = output_stat->st_dev;
		out_ino = output_stat->st_ino;
		check_redirection = 1;
	}
	else
		check_redirection = 0;

	infile = (char*)&"-";
	argind = optind;

	/*
	 * allocate an initial amount of line structures; we allocate these
	 * outside of the main loop (and also free them outside of the main
	 * loop, see below), because they can be reused by possible subsequent
	 * files.
	 */
	lines = lines_alloc( lines, 0, lines_allocated=DEFAULT_LINES );

	/* main loop */
	do
	{
		/*
		 * check if any of the input files are
		 * the same as the output file.
		 */

		if (argc>argind)
			infile = argv[argind];

		if (verbose)
			info( "file: %s\n", infile );

		if (is_stdin(infile))
		{
			have_read_stdin = 1;
			input_desc = STDIN;
		}
		else
		{
			input_desc = open( infile, O_RDONLY );
			if (input_desc<0)
			{
				error( CONT, errno, "%s", infile );
				continue;
			}
		}

		if (fstat( input_desc, input_stat )<0)
		{
			error( CONT, errno, "%s", infile );
			goto contin;
		}

		/*
		 * compare the device and i-node numbers of this input file with
		 * the corresponding values of the (output file associated with)
		 * stdout, and skip this input file if they coincide.  Input
		 * files cannot be redirected to themselves.
		 */
		if (check_redirection && input_stat->st_dev==out_dev && input_stat->st_ino==out_ino)
		{
			error( CONT, 0, "skipping %s: input file is output file", infile );
			goto contin;
		}

		input_stream = fdopen( input_desc, "r" );

		/*
		 * here's where we call the functions that get most
		 * of the work done.
		 */
		read_file( input_stream );
		tilt_file();
		clear_lines();

		contin:

		if (!is_stdin(infile) && fclose( input_stream )<0)
			error( CONT, errno, "%s", infile );
	}
	 while (++argind<argc);

	if (have_read_stdin && close( STDIN )<0)
		error( EXIT, errno, "-" );

	if (close( STDOUT )<0)
		error( EXIT, errno, "write error" );

	/*
	 * clean up this one (clear_lines called above already has
	 * freed all lines[].str pointers.
	 */
	if (lines!=NULL)
		free( lines );
}

/*---------------------------------------------------------------------------*/
